CodeMirror.defineMode("sql", function(config, parserConfig) {
    "use strict";

    var client = parserConfig.client || {},
            atoms = parserConfig.atoms || {"false": true, "true": true, "null": true},
    builtin = parserConfig.builtin || {},
            keywords = parserConfig.keywords || {},
            operatorChars = parserConfig.operatorChars || /^[*+\-%<>!=&|~^]/,
            support = parserConfig.support || {},
            hooks = parserConfig.hooks || {},
            dateSQL = parserConfig.dateSQL || {"date": true, "time": true, "timestamp": true};

    function tokenBase(stream, state) {
        var ch = stream.next();

        // call hooks from the mime type
        if (hooks[ch]) {
            var result = hooks[ch](stream, state);
            if (result !== false)
                return result;
        }
        if (ch == "/") {
            if (stream.eat("*")) {
                state.tokenize = tokenComment;
                return tokenComment(stream, state);
            }
            if (stream.eat("/")) {
                stream.skipToEnd();
                return "comment";
            }
        }
        if (support.hexNumber == true &&
                ((ch == "0" && stream.match(/^[xX][0-9a-fA-F]+/))
                        || (ch == "x" || ch == "X") && stream.match(/^'[0-9a-fA-F]+'/))) {
            // hex
            // ref: http://dev.mysql.com/doc/refman/5.5/en/hexadecimal-literals.html
            return "number";
        } else if (support.binaryNumber == true &&
                (((ch == "b" || ch == "B") && stream.match(/^'[01]+'/))
                        || (ch == "0" && stream.match(/^b[01]+/)))) {
            // bitstring
            // ref: http://dev.mysql.com/doc/refman/5.5/en/bit-field-literals.html
            return "number";
        } else if (ch.charCodeAt(0) > 47 && ch.charCodeAt(0) < 58) {
            // numbers
            // ref: http://dev.mysql.com/doc/refman/5.5/en/number-literals.html
            stream.match(/^[0-9]*\.?[0-9]+([eE][-+]?[0-9]+)?/);
            support.decimallessFloat == true && stream.eat('.');
            return "number";
        } else if (ch == "?" && (stream.eatSpace() || stream.eol() || stream.eat(";"))) {
            // placeholders
            return "variable-3";
        } else if (ch == "'" || (ch == '"' && support.doubleQuote)) {
            // strings
            // ref: http://dev.mysql.com/doc/refman/5.5/en/string-literals.html
            state.tokenize = tokenLiteral(ch);
            return state.tokenize(stream, state);
        } else if ((((support.nCharCast == true && (ch == "n" || ch == "N"))
                || (support.charsetCast == true && ch == "_" && stream.match(/[a-z][a-z0-9]*/i)))
                && (stream.peek() == "'" || stream.peek() == '"'))) {
            // charset casting: _utf8'str', N'str', n'str'
            // ref: http://dev.mysql.com/doc/refman/5.5/en/string-literals.html
            return "keyword";
        } else if (/^[\(\),\;\[\]]/.test(ch)) {
            // no highlightning
            return null;
        } else if (support.commentSlashSlash && ch == "/" && stream.eat("/")) {
            // 1-line comment
            stream.skipToEnd();
            return "comment";
        } else if ((support.commentHash && ch == "#")
                || (ch == "-" && stream.eat("-") && (!support.commentSpaceRequired || stream.eat(" ")))) {
            // 1-line comments
            // ref: https://kb.askmonty.org/en/comment-syntax/
            stream.skipToEnd();
            return "comment";
        } else if (ch == "/" && stream.eat("*")) {
            // multi-line comments
            // ref: https://kb.askmonty.org/en/comment-syntax/
            state.tokenize = tokenComment;
            return state.tokenize(stream, state);
        } else if (ch == ".") {
            // .1 for 0.1
            if (support.zerolessFloat == true && stream.match(/^(?:\d+(?:e[+-]?\d+)?)/i)) {
                return "number";
            }
            // .table_name (ODBC)
            // // ref: http://dev.mysql.com/doc/refman/5.6/en/identifier-qualifiers.html
            if (support.ODBCdotTable == true && stream.match(/^[a-zA-Z_]+/)) {
                return "variable-2";
            }
        } else if (operatorChars.test(ch)) {
            // operators
            stream.eatWhile(operatorChars);
            return null;
        } else if (ch == '{' &&
                (stream.match(/^( )*(d|D|t|T|ts|TS)( )*'[^']*'( )*}/) || stream.match(/^( )*(d|D|t|T|ts|TS)( )*"[^"]*"( )*}/))) {
            // dates (weird ODBC syntax)
            // ref: http://dev.mysql.com/doc/refman/5.5/en/date-and-time-literals.html
            return "number";
        } else {
            stream.eatWhile(/^[_\w\d]/);
            var word = stream.current().toLowerCase();
            // dates (standard SQL syntax)
            // ref: http://dev.mysql.com/doc/refman/5.5/en/date-and-time-literals.html
            if (dateSQL.hasOwnProperty(word) && (stream.match(/^( )+'[^']*'/) || stream.match(/^( )+"[^"]*"/)))
                return "number";
            if (atoms.hasOwnProperty(word))
                return "atom";
            if (builtin.hasOwnProperty(word))
                return "builtin";
            if (keywords.hasOwnProperty(word))
                return "keyword";
            if (client.hasOwnProperty(word))
                return "string-2";
            return null;
        }
    }

    // 'string', with char specified in quote escaped by '\'
    function tokenLiteral(quote) {
        return function(stream, state) {
            var escaped = false, ch;
            while ((ch = stream.next()) != null) {
                if (ch == quote && !escaped) {
                    state.tokenize = tokenBase;
                    break;
                }
                escaped = !escaped && ch == "\\";
            }
            return "string";
        };
    }
    function tokenComment(stream, state) {
        while (true) {
            if (stream.skipTo("*")) {
                stream.next();
                if (stream.eat("/")) {
                    state.tokenize = tokenBase;
                    break;
                }
            } else {
                stream.skipToEnd();
                break;
            }
        }
        return "comment";
    }

    function pushContext(stream, state, type) {
        state.context = {
            prev: state.context,
            indent: stream.indentation(),
            col: stream.column(),
            type: type
        };
    }

    function popContext(state) {
        state.indent = state.context.indent;
        state.context = state.context.prev;
    }

    return {
        startState: function() {
            return {tokenize: tokenBase, context: null};
        },
        token: function(stream, state) {
            if (stream.sol()) {
                if (state.context && state.context.align == null)
                    state.context.align = false;
            }
            if (stream.eatSpace())
                return null;

            var style = state.tokenize(stream, state);
            if (style == "comment")
                return style;

            if (state.context && state.context.align == null)
                state.context.align = true;

            var tok = stream.current();
            if (tok == "(")
                pushContext(stream, state, ")");
            else if (tok == "{")
                pushContext(stream, state, "}");
            else if (state.context && state.context.type == tok)
                popContext(state);
            return style;
        },
        indent: function(state, textAfter) {
            var cx = state.context;
            if (!cx)
                return CodeMirror.Pass;
            if (cx.align)
                return cx.col + (textAfter.charAt(0) == cx.type ? 0 : 1);
            else
                return cx.indent + config.indentUnit;
        },
        lineComment: "//"
    };
});

(function() {
    "use strict";

    var sqlKeywords = "create data drop for of and insert into kill learnmodel learnmodels list neuralnetwork neuralnetworks run show status train values with";

    function set(str) {
        var obj = {}, words = str.split(" ");
        for (var i = 0; i < words.length; ++i)
            obj[words[i]] = true;
        return obj;
    }

    CodeMirror.defineMIME("text/x-annsql", {
        name: "sql",
        keywords: set(sqlKeywords),
        builtin: set("boolean integer real string date"),
        atoms: set("false true null"),
        operatorChars: /^[*+\-%<>!=]/,
        support: set("ODBCdotTable doubleQuote binaryNumber hexNumber")
    });
}());
